from myrrh.utils.mstring import str2int
from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.framework.mpython._mosenv import AbcOsEnv

__mlib__ = "OsEnv"


class OsEnv(AbcOsEnv):
    name = "nt"
    linesep = "\r\n"

    def putenv(self, varname, value):
        _, err, rval = self.myrrh_os.cmdb(
            b'%(setx)s "%(varname)s" "%(value)s"',
            varname=self.myrrh_os.sh_escape_bytes(varname),
            value=self.myrrh_os.sh_escape_bytes(value),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def unsetenv(self, varname):
        _, err, rval = self.myrrh_os.cmdb(
            rb'%(reg)s DELETE HKCU\Environment /F /V "%(varname)s"',
            varname=self.myrrh_os.sh_escape_bytes(varname),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def gethome(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s %%HOMEDRIVE%%%%HOMEPATH%%")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def gettemp(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s %%TEMP%%")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def getlogin(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s %%USERNAME%%@%%USERDOMAIN%%")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def geteuid(self):
        return 0

    def getegid(self):
        return 0

    def getuid(self):
        return 0

    def getgid(self):
        return 0

    def getgroups(self):
        return [0]

    def cpu_count(self):
        out, err, rval = self.myrrh_os.cmd(b"%(wmic)s cpu get NumberOfCores")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        out = out.split("\r\n")
        return str2int(out[1])



import threading

from myrrh import factory
from myrrh.core.services import groups
from myrrh.framework import Runtime

from ._bmy_exceptions import BmyNotReady


class BmyEntity:
    def __init__(self, assembly: factory.Assembly, eid: str):
        self._assembly = assembly
        self._eid = eid
        self._entity = None
        self._runtime = None
        self._used: list[int] = list()

    def __str__(self) -> str:
        return self._eid

    def build(self):
        if not self._entity:
            self._entity = self._assembly.build()

        return self._entity

    @property
    def eid(self):
        return self._eid

    @property
    def cfg(self):
        if not self.built:
            raise BmyNotReady(self._eid)

        return self._entity.cfg

    @property
    def system(self):
        if not self.built:
            raise BmyNotReady(self.eid)

        return self._entity.system

    @property
    def host(self):
        if not self.built:
            raise BmyNotReady(self.eid)

        return self._entity.host

    @property
    def vendor(self):
        if not self.built:
            raise BmyNotReady(self.eid)

        return self._entity.vendor

    @property
    def runtime(self):
        if not self._runtime:
            self._runtime = Runtime(self._entity.system)

        return self._runtime

    @groups.myrrh_group_sync_member
    def release(self):
        try:
            self._used.remove(threading.get_ident())
        except ValueError:
            pass

    @groups.myrrh_group_sync_member
    def acquire(self):
        self._used.append(threading.get_ident())

    @property
    def built(self):
        return self._entity is not None

    @property
    def used(self):
        return len(self._used) != 0



import typing
import pydantic

from myrrh.utils import secrets

from .item import BaseItem


class Credential(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")

    login: str = pydantic.Field()
    email: str | None = None
    domain: str | None = None
    role: str | None = None
    password: str | None = None
    private_key: str = ""

    @pydantic.field_validator("password")
    def password_validator(cls, password, info):
        return secrets.encrypt(password)


class Credentials(BaseItem[typing.Literal["credentials"]]):
    """
    Credentials attach to an entity.
    """

    credentials: list[Credential] = pydantic.Field(default_factory=list)



import unittest
import threading

from myrrh.core.services.system import Buffer


class TestBuffer(unittest.TestCase):
    def test_basic_buffer_create(self):
        b = Buffer(threading.RLock(), threading.RLock(), 100)

        self.assertFalse(b.closed)
        self.assertEqual(b.wr_size, 100)

    def test_basic_buffer_write_read(self):
        b = Buffer(threading.RLock(), threading.RLock(), 100)

        b.write(b"01234")
        self.assertEqual(b.rd_size, 5)
        self.assertEqual(b.rd_pos, 0)

        d = b.read()
        self.assertEqual(d, b"01234")
        self.assertEqual(b.rd_pos, 5)
        self.assertEqual(b.rd_size, 0)

        b.write(b"56789")
        self.assertEqual(b.rd_size, 5)
        self.assertEqual(b.rd_pos, 5)

        d = b.read(2)
        self.assertEqual(d, b"56")
        self.assertEqual(b.rd_pos, 7)
        self.assertEqual(b.rd_size, 3)

        d = b.read()
        self.assertEqual(d, b"789")
        self.assertEqual(b.rd_pos, 10)
        self.assertEqual(b.rd_size, 0)

    def test_basic_buffer_n_write_read(self):
        b = Buffer(threading.RLock(), threading.RLock(), 100)

        b.write(b"01234")
        self.assertEqual(b.rd_size, 5)
        self.assertEqual(b.rd_pos, 0)

        b.write(b"56789")
        self.assertEqual(b.rd_size, 10)
        self.assertEqual(b.rd_pos, 0)

        d = b.read()
        self.assertEqual(d, b"0123456789")
        self.assertEqual(b.rd_pos, 10)
        self.assertEqual(b.rd_size, 0)

    def test_basic_rotate(self):
        b = Buffer(threading.RLock(), threading.RLock(), 100)

        for i in range(0, 10):
            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            b.write(b"0" * 90)
            self.assertEqual(b.rd_pos, prev_rd_pos)
            self.assertEqual(b.wr_pos, (prev_wr_pos + 90) % 100)
            self.assertEqual(b.rd_size, 90)
            self.assertEqual(b.wr_size, 10)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            d = b.read()
            self.assertEqual(d, b"0" * 90)
            self.assertEqual(b.rd_pos, (prev_rd_pos + 90) % 100)
            self.assertEqual(b.wr_pos, prev_wr_pos)
            self.assertEqual(b.rd_size, 0)
            self.assertEqual(b.wr_size, 100)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            b.write(b"1" * 90)
            self.assertEqual(b.rd_pos, prev_rd_pos)
            self.assertEqual(b.wr_pos, (prev_wr_pos + 90) % 100)
            self.assertEqual(b.rd_size, 90)
            self.assertEqual(b.wr_size, 10)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            d = b.read(50)
            self.assertEqual(d, b"1" * 50)
            self.assertEqual(b.rd_pos, (prev_rd_pos + 50) % 100)
            self.assertEqual(b.wr_pos, prev_wr_pos)
            self.assertEqual(b.rd_size, 40)
            self.assertEqual(b.wr_size, 60)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            b.write(b"2" * 20)
            self.assertEqual(b.rd_pos, prev_rd_pos)
            self.assertEqual(b.wr_pos, (prev_wr_pos + 20) % 100)
            self.assertEqual(b.rd_size, 60)
            self.assertEqual(b.wr_size, 40)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            d = b.read(40)
            self.assertEqual(d, b"1" * 40)
            self.assertEqual(b.wr_pos, prev_wr_pos)
            self.assertEqual(b.rd_pos, (prev_rd_pos + 40) % 100)
            self.assertEqual(b.rd_size, 20)
            self.assertEqual(b.wr_size, 80)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            d = b.read()
            self.assertEqual(d, b"2" * 20)
            self.assertEqual(b.rd_pos, (prev_rd_pos + 20) % 100)
            self.assertEqual(b.rd_size, 0)
            self.assertEqual(b.wr_size, 100)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            b.write(b"3" * 30)
            self.assertEqual(b.wr_pos, (prev_wr_pos + 30) % 100)
            self.assertEqual(b.rd_pos, prev_rd_pos)
            self.assertEqual(b.rd_size, 30)
            self.assertEqual(b.wr_size, 70)

            prev_rd_pos = b.rd_pos
            prev_wr_pos = b.wr_pos
            d = b.read()
            self.assertEqual(d, b"3" * 30)
            self.assertEqual(b.wr_pos, prev_wr_pos)
            self.assertEqual(b.rd_pos, (prev_rd_pos + 30) % 100)
            self.assertEqual(b.rd_size, 0)
            self.assertEqual(b.wr_size, 100)


if __name__ == "__main__":
    unittest.main()



import bmy

from .exceptions import Reboot
from prompt_toolkit.application import get_app

from prompt_toolkit.filters import (
    Condition,
    has_focus,
)

from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys


def load_confirm_exit_bindings(prompt):
    """
    Handle yes/no key presses when the exit confirmation is shown.
    """
    bindings = KeyBindings()

    handle = bindings.add
    confirmation_visible = Condition(lambda: prompt.show_exit_confirmation)

    @handle("y", filter=confirmation_visible)
    @handle("Y", filter=confirmation_visible)
    @handle("enter", filter=confirmation_visible)
    @handle("c-d", filter=confirmation_visible)
    def _(event):
        """
        Really quit.
        """
        if prompt.reboot_required:
            event.app.exit(exception=Reboot, style="class:exiting")
        else:
            event.app.exit(exception=EOFError, style="class:exiting")

    @handle(Keys.Any, filter=confirmation_visible)
    def _(event):
        """
        Cancel exit.
        """
        prompt.show_exit_confirmation = False
        prompt.reboot_required = False
        prompt.app.layout.focus_previous()

    return bindings


def load_prompt_bindings(prompt):
    """
    Custom key bindings.
    """
    bindings = KeyBindings()

    handle = bindings.add

    @handle(
        "c-r",
        filter=has_focus(prompt.default_buffer)
        & Condition(
            lambda:
            # The current buffer is empty.
            not get_app().current_buffer.text
        ),
    )
    def _(event):
        """
        Override Control-D exit, to ask for confirmation.
        """
        if prompt.confirm_exit:
            # Show exit confirmation and focus it (focusing is important for
            # making sure the default buffer key bindings are not active).
            prompt.show_exit_confirmation = True
            prompt.reboot_required = True
            prompt.app.layout.focus(prompt.exit_confirmation)
        else:
            event.app.exit(exception=Reboot)

    @handle(
        "c-d",
        filter=has_focus(prompt.default_buffer)
        & Condition(
            lambda:
            # The current buffer is empty.
            not get_app().current_buffer.text
        ),
    )
    def _(event):
        """
        Override Control-D exit, to ask for confirmation.
        """
        if prompt.confirm_exit:
            # Show exit confirmation and focus it (focusing is important for
            # making sure the default buffer key bindings are not active).
            prompt.show_exit_confirmation = True
            prompt.reboot_required = False
            prompt.app.layout.focus(prompt.exit_confirmation)
        else:
            event.app.exit(exception=EOFError)

    @handle("c-c", filter=has_focus(prompt.default_buffer))
    def _(event):
        "Abort when Control-C has been pressed."
        event.app.exit(exception=KeyboardInterrupt, style="class:aborting")

    @handle("c-z")
    def _(event):
        bmy.next()

    @handle("c-a")
    def _(event):
        bmy.previous()

    return bindings



import os
import sys
import platform
import locale
import uuid
import tempfile
import typing

import myrrh

import myrrh.warehouse
import myrrh.provider


from .system import Shell, Stream, FileSystem, StreamSystemAPI

__version__ = myrrh.__version__

WINDOWS = sys.platform == "win32"

if WINDOWS:
    import ctypes
    from ctypes import windll  # type: ignore[attr-defined]

    default_shell_encoding = "cp%d" % windll.kernel32.GetConsoleOutputCP()

    def GetUserName(format):
        buf = None
        sz = ctypes.pointer(ctypes.c_ulong(0))
        try:
            ctypes.windll.secur32.GetUserNameExW(format, buf, sz)
            buf = ctypes.create_unicode_buffer(sz.contents.value)
            if windll.secur32.GetUserNameExW(format, buf, sz):
                return buf.value
        finally:
            if buf is not None:
                del buf
            del sz

    username = GetUserName(2) or os.getlogin()
    uid = GetUserName(6)
    gid = None

else:
    default_shell_encoding = "utf-8"
    try:
        username = os.getlogin()
        uid = os.geteuid()  # type: ignore[attr-defined]
        gid = os.getegid()  # type: ignore[attr-defined]
    except OSError:
        username = "unknown"
        uid = -1
        gid = -1


class LocalProviderSettings(myrrh.warehouse.Settings):
    name: typing.Literal["local"]
    cwd: str | None = None


class Provider(myrrh.provider.IProvider):
    """
    Local tools commands
    """

    def __init__(self, settings=None):
        """
        instantiate specific local services
        """
        if settings and settings.cwd:
            cwd = os.path.expandvars(settings.cwd)
            cwd = os.path.expanduser(cwd)
            os.chdir(cwd)

    def services(self):
        return (Stream, Shell, FileSystem, StreamSystemAPI)

    def catalog(self):
        return ("system", "id")

    def deliver(self, name: str) -> dict:
        match name:
            case "system":
                cwd = os.getcwd()
                return myrrh.warehouse.System(  # type: ignore[call-arg]
                    os=os.name,
                    label=platform.node(),
                    description=":".join(platform.uname()),
                    fsencoding=sys.getfilesystemencoding(),
                    localecode=locale.getlocale()[0] or "LC_CTYPE",
                    encoding=locale.getencoding(),
                    default_shell_encoding=default_shell_encoding,
                    shell=os.environ.get("COMSPEC", "cmd.exe") if WINDOWS else "/bin/sh",
                    shellargs=("/c",) if WINDOWS else ("-c",),
                    cwd=cwd,
                    username=username,
                    uid=uid,
                    gid=gid,
                    tmpdir=tempfile.gettempdir(),
                    env=dict(os.environ),
                ).model_dump()

            case "id":
                return myrrh.warehouse.Id(id=platform.node(), uuid=uuid.uuid1()).model_dump()

            case _:
                return dict()


provider = Provider
provider_settings = LocalProviderSettings



import contextlib
import sys
import threading
import importlib
import re

from myrrh.core.services.system import _mlib_

_lock = threading.RLock()


def _modname(name):
    return re.sub("(_*)(.+)", r"\g<1>m\g<2>", name)


class MLibModule(threading.local):
    __slots__ = ["__package__", "__name__", "py", "_mlib_mod"]

    __path__ = "(Mlib)"
    __loader__ = __loader__  # type: ignore[name-defined]
    __doc__ = __doc__
    __file__ = __file__
    __spec__ = __spec__  # type: ignore[name-defined]

    def __init__(self, name, *, module=None, package="", parent=None):
        self.__package__ = package
        self.__name__ = name
        self._mlib_mod = module

        if not parent:
            self._mlib_systems = list()
        else:
            self._mlib_systems = parent._mlib_systems

    def _mlib_cls(self, s, *, func=lambda mod: mod):
        if self._mlib_mod is None:
            return MLibModule
        return _mlib_(self._mlib_mod, lambda s: func(self._mlib_mod))(s)

    def __repr__(self):
        return f"<MlibModule> {repr(_mlib_(self._mlib_mod, self._mlib_mod))}"

    def __setattr__(self, name, value):
        with _lock:
            if name in (
                "__package__",
                "__name__",
                "__spec__",
                "__loader__",
                "__file__",
                "__path__",
                "_mlib_mod",
                "_mlib_systems",
                "__finder__",
            ):
                return super().__setattr__(name, value)

            if not self._mlib_mod:
                return super().__setattr__(name, value)

            setattr(
                _mlib_(self._mlib_mod, self._mlib_mod),
                name,
                getattr(value, "_mlib_mod", value),
            )

    def __getattr__(self, name):
        with _lock:
            if self._mlib_systems:
                system = self.mlib_current()
                try:
                    mod = self._mlib_cls(system)
                except (NotImplementedError, AttributeError, TypeError) as e:
                    raise ImportError(f"invalid module {repr(self._mlib_mod)} : {e}")

                result = getattr(mod, name)

                try:
                    return _mlib_(result, lambda s: result)(system)

                except (NotImplementedError, AttributeError, TypeError) as e:
                    raise ImportError(f"invalid module {repr(result)} : {e}")

            mod = _mlib_(self._mlib_mod, self._mlib_mod)
            mod = getattr(mod, name)
            return _mlib_(mod, mod)

    @contextlib.contextmanager
    def mlib_select(self, system):
        try:
            self.mlib_push(system)
            yield system
        finally:
            self.mlib_pop()

    def mlib_push(self, system):
        self._mlib_systems.append(system)

    def mlib_pop(self):
        self._mlib_systems.pop()

    def mlib_current(self):
        if self._mlib_systems:
            return self._mlib_systems[-1]


class MLibModuleLoader:
    def __init__(self, name, spec, *, package="", parent=None):
        self.module_name = name
        self.module_spec = spec
        self.module_package = package
        self.module_parent = parent

    def create_module(self, spec):
        _lock.acquire()
        try:
            module = sys.modules.get(self.module_spec.name)
            spec.loader_state = not module
            if not module:
                module = importlib.util.module_from_spec(self.module_spec)
            else:
                module.__spec__.loader_state = True

            module = MLibModule(
                self.module_name,
                module=module,
                parent=self.module_parent,
                package=self.module_package,
            )
            module.__spec__ = spec

            return module
        except:  # noqa: E722
            _lock.release()
            raise

    def exec_module(self, module):
        try:
            if module.__spec__.loader_state:
                mlib_mod = module._mlib_mod
                mlib_mod.__loader__.exec_module(mlib_mod)
                sys.modules[mlib_mod.__spec__.name] = mlib_mod
        finally:
            _lock.release()


class MLibSubModuleFinder:
    packages: dict[str, str] = dict()

    def find_spec(self, fullname, path, target=None):
        package, _, name = fullname.rpartition(".")

        if package not in (__module__.__name__, __name__):
            return

        if name in self.packages:
            module_parent = sys.modules[package]
            module_path = self.packages[name]
            spec = importlib.util.find_spec(module_path)

            if spec is not None:
                module_spec = importlib.machinery.ModuleSpec(
                    fullname,
                    MLibModuleLoader(fullname, spec, package=module_path, parent=module_parent),
                )
                return module_spec


class MLibModuleFinder:
    def find_spec(self, fullname, path, target=None):
        if fullname.startswith((__module__.__name__, __name__)):
            try:
                package, _, name = fullname.rpartition(".")
                module_parent = sys.modules[package]

                if isinstance(module_parent, MLibModule):
                    parent_package = module_parent.__package__
                    module_name = _modname(name)
                    module_package = ".".join((parent_package, module_name))

                    spec = importlib.util.find_spec(module_package)

                    if spec is not None:
                        module_spec = importlib.machinery.ModuleSpec(
                            fullname,
                            MLibModuleLoader(
                                fullname,
                                spec,
                                package=module_package,
                                parent=module_parent,
                            ),
                        )
                        return module_spec

                    arch = None
                    current = __module__.mlib_current()
                    module_arch_package = None

                    if current is not None:
                        try:
                            arch = current.myrrh_os.cfg.system.os
                        except Exception:
                            ...

                    if arch:
                        parent_package = "myrrh.framework.arch." + arch + parent_package.removeprefix("myrrh.framework")
                        module_arch_package = ".".join((parent_package, module_name))

                    if module_arch_package:
                        try:
                            spec = importlib.util.find_spec(module_arch_package)
                        except ModuleNotFoundError:
                            pass

                    if spec is not None:
                        module_spec = importlib.machinery.ModuleSpec(
                            fullname,
                            MLibModuleLoader(
                                fullname,
                                spec,
                                package=module_arch_package,
                                parent=module_parent,
                            ),
                        )
                        return module_spec

                    # python special case
                    if fullname.startswith("mlib.py."):
                        module_package = fullname[len("mlib.py.") :]
                        spec = importlib.util.find_spec(module_package)
                        if spec is not None:
                            return importlib.machinery.ModuleSpec(
                                fullname,
                                MLibModuleLoader(
                                    fullname,
                                    spec,
                                    package=module_package,
                                    parent=module_parent,
                                ),
                            )

            except KeyError:
                pass


__subfinder__ = MLibSubModuleFinder()
__finder__ = MLibModuleFinder()


__module__ = MLibModule("mlib", package="myrrh.framework.mlib")
__module__.__finder__ = __subfinder__

sys.modules["mlib"] = __module__  # type: ignore[assignment]
sys.modules[__name__] = __module__  # type: ignore[assignment]

sys.meta_path.append(__subfinder__)
sys.meta_path.append(__finder__)



import unittest
import logging
import os

import bmy

from myrrh.core.services import log
from myrrh.utils.myrrh_test_init import setUp

setUp(f"bmy.{os.path.basename(__file__)}")


class Test_basic(unittest.TestCase):
    def test_info(self):
        i = bmy.info()
        for k in (
            "id",
            "description",
            "location",
            "os",
            "services",
            "catalog",
            "cwd",
            "warehouse",
        ):
            self.assertIn(k, i)
        i = bmy.info("id.id")
        self.assertIsInstance(i, str)
        i = bmy.info("notvalid.wiz")
        self.assertEqual(str(i), "na")

    def test_debug(self):
        current_level = bmy.debug()
        try:
            bmy.debug("DEBUG")
            self.assertEqual(log.level, logging.DEBUG)
            bmy.debug("INFO")
            self.assertEqual(log.level, logging.INFO)
            bmy.debug(logging.CRITICAL)
            self.assertEqual(log.level, logging.CRITICAL)
        finally:
            bmy.debug(current_level)

    def test_save_load(self):
        import os

        try:
            bmy.save("test_entity.emyrrh")
            self.assertTrue(os.path.isfile("test_entity.emyrrh"))
            bmy.load("test_entity.emyrrh")

            bmy.save("test_entity_full.emyrrh", full=True)
            e2 = bmy.load("test_entity_full.emyrrh")
            bmy.build(eid=e2)
            self.assertEqual(bmy.info("warehouse.id.uuid", eid=e2), bmy.info("warehouse.id.uuid"))
            self.assertTrue(os.path.isfile("test_entity_full.emyrrh"))

        finally:
            os.remove("test_entity.emyrrh")
            os.remove("test_entity_full.emyrrh")

    def test_which(self):
        cmd = bmy.info("warehouse.system.shell")

        p = bmy.which(bmy.info("warehouse.system.shell"))

        try:
            p.index(cmd)
        except ValueError:
            self.fail("{cmd} not found in path returns by which")

    def test_launch_kill(self):
        pid = bmy.launch("ping 127.0.0.1")
        bmy.kill(pid)


if __name__ == "__main__":
    unittest.main()



from myrrh.framework.mfs.madvfs import AbcAdvFs
from myrrh.core.services.system import ExecutionFailureCauseRVal

__mlib__ = "AdvFs"


class AdvFs(AbcAdvFs):
    default_mode = 0o777

    CHUNK_HEADER_SCRIPT = b"""\
path=$0
exec 0<"$path"
while IFS= read -r f; do
case $f in
"exit")
echo exit
break;
esac
done
read szs
read files
splitter()
{
IFS=:
set $files
for s in $szs ; do
dd of="$1" count=$s iflag=count_bytes
shift
done
}
splitter
rm -f "$path"
exit"""

    def chown(self, path, user, group):
        path = self.myrrh_os.shencode(path)
        user = self.myrrh_os.shencode(user)
        group = self.myrrh_os.shencode(group)
        _, _, rval = self.myrrh_os.shell.execute(b"/bin/chown %s:%s %s" % (user, group, path))
        return rval == 0

    def chmod(self, mode, path):
        path = self.myrrh_os.shencode(path)
        _, _, rval = self.myrrh_os.shell.execute(b"/bin/chmod %o %s" % (mode, path))
        return rval == 0

    def _localtransfer(self, srcs, dest):
        srcs = [b"'%s'" % src for src in srcs]
        srcs = b" ".join(srcs)

        _, err, rval = self.myrrh_os.cmdb(
            b"%(cp)s -a -f %(srcs)s %(dest)s",
            srcs=srcs,
            dest=dest and self.myrrh_os.sh_escape_bytes(dest) or b".",
        )
        if b"non-directory" in err:
            err = b"Not a directory"

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _localmove(self, srcs, dest):
        if len(srcs) > 1:
            srcs = b" ".join((self.myrrh_os.sh_escape_bytes(s) for s in srcs))
            _, err, rval = self.myrrh_os.cmdb(
                b"if [ -d '%(dest)s' ] ; then %(mv)s %(srcs)s '%(dest)s'; else exit 10; fi",
                srcs=srcs,
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )
        else:
            _, err, rval = self.myrrh_os.cmdb(
                b"%(mv)s %(srcs)s %(dest)s",
                srcs=self.myrrh_os.sh_escape_bytes(srcs[0]),
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _localremove(self, srcs):
        srcs = b" ".join((self.myrrh_os.sh_escape_bytes(s) for s in srcs))
        _, err, rval = self.myrrh_os.cmdb(b"%(rm)s -Rf %(srcs)s", srcs=srcs)

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _chunk_header(self, file_descs, *, destdir=None):
        szs = (str(f[2]).encode() for f in file_descs)
        paths = (self.myrrh_os.p(f[1]) for f in file_descs)
        return b"\n".join((self.CHUNK_HEADER_SCRIPT, b":".join(szs), b":".join(paths))) + b"\n"

    def _unchunk(self, path):
        path = self.myrrh_os.p(path)
        _, e, r = self.myrrh_os.cmdb(b"%(sh)s %(path)s", path=self.myrrh_os.sh_escape_bytes(path))
        ExecutionFailureCauseRVal(self, e, r, 0).check()

    def _write_chunk(self, path, file_descs):
        path = self.myrrh_os.p(path)
        files = b" ".join((b"%s" % self.myrrh_os.sh_escape_bytes(self.myrrh_os.p(file)) for file, _, _ in file_descs))
        _, e, r = self.myrrh_os.cmdb(
            b'%(cat)s %(files)s > "%(path)s"',
            files=files,
            path=self.myrrh_os.sh_escape_bytes(path),
        )
        ExecutionFailureCauseRVal(self, e, r, 0).check()

    def _mkdirs(self, dirs: list):
        cmd = b'for d in %(dirs)s; do %(mkdir)s -p "$d"; done'

        dirs: bytes = b" ".join(dirs)
        _, e, r = self.myrrh_os.cmdb(cmd, dirs=dirs)
        ExecutionFailureCauseRVal(self, e, r, 0).check()

    def _scanfiles(self, files):
        """
        return files(path, size)
        """
        result = []
        for f in files:
            stat = self.myrrh_os.fs.stat(f)
            result.append((f, stat.st_size))

        return result

    def _scantree(self, path):
        dirs = []
        files = []

        def _loop(startpath=None):
            paths = self.myrrh_os.fs.list(startpath or path)
            for p, i in paths:
                p = self.myrrh_os.joinpath(paths.cname, p)
                if paths.is_container(i):
                    dirs.append(p)
                    _loop(p)
                else:
                    files.append((p, paths.size(i)))

        _loop()

        return dirs, files



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins
from . import mimportlib

__mlib__ = "AbcPlatform"


class _interface(ABC):
    import platform as local_platform

    @property
    @abstractmethod
    def uname_result(self) -> local_platform.uname_result:
        ...

    @property
    @abstractmethod
    def _ver_output(self):
        ...

    @property
    @abstractmethod
    def _WIN32_CLIENT_RELEASES(self):
        ...

    @property
    @abstractmethod
    def _WIN32_SERVER_RELEASES(self):
        ...

    @property
    @abstractmethod
    def _Processor(self):
        ...

    @property
    @abstractmethod
    def _sys_version(self):
        ...

    @property
    @abstractmethod
    def _sys_version_cache(self):
        ...

    @property
    @abstractmethod
    def _os_release_candidates(self):
        ...

    @abstractmethod
    def win32_is_iot(self) -> bool:
        ...

    @abstractmethod
    def win32_edition(self) -> str:
        ...

    @abstractmethod
    def win32_ver(self, release: str = ..., version: str = ..., csd: str = ..., ptype: str = ...) -> tuple[str, str, str, str]:
        ...

    @abstractmethod
    def mac_ver(self, release: str = ..., version: str = ..., csd: str = ..., ptype: str = ...) -> tuple[str, str, str, str]:
        ...

    @abstractmethod
    def java_ver(
        self,
        release: str = ...,
        vendor: str = ...,
        vminfo: tuple[str, str, str] = ...,
        osinfo: tuple[str, str, str] = ...,
    ) -> tuple[str, str, tuple[str, str, str], tuple[str, str, str]]:
        ...

    @abstractmethod
    def architecture(self, executable: str = ..., bits: str = ..., linkage: str = ...) -> tuple[str, str]:
        ...

    @abstractmethod
    def uname(self) -> str:
        ...

    @abstractmethod
    def system(self) -> str:
        ...

    @abstractmethod
    def node(self) -> str:
        ...

    @abstractmethod
    def release(self) -> str:
        ...

    @abstractmethod
    def version(self) -> str:
        ...

    @abstractmethod
    def machine(self) -> str:
        ...

    @abstractmethod
    def processor(self) -> str:
        ...

    @abstractmethod
    def python_implementation(self) -> str:
        ...

    @abstractmethod
    def python_version(self) -> str:
        ...

    @abstractmethod
    def python_version_tuple(self) -> tuple[str, str, str]:
        ...

    @abstractmethod
    def python_branch(self) -> str:
        ...

    @abstractmethod
    def python_revision(self) -> str:
        ...

    @abstractmethod
    def python_build(self) -> tuple[str, str]:
        ...

    @abstractmethod
    def python_compiler(self) -> str:
        ...

    @abstractmethod
    def platform(self, aliased: bool = ..., terse: bool = ...) -> str:
        ...

    @abstractmethod
    def freedesktop_os_release(self) -> dict[str, str]:
        ...

    @abstractmethod
    def system_alias(
        self, system: typing.Any, release: typing.Any, version: typing.Any
    ) -> tuple | tuple[typing.Any | typing.Literal["Solaris", "Windows"], str | typing.Any, typing.Any,]:
        ...

    @abstractmethod
    def libc_ver(
        self,
        executable: typing.Any | None = None,
        lib: str = "",
        version: str = "",
        chunksize: int = 16384,
    ) -> typing.Any:
        ...

    @abstractmethod
    def _parse_os_release(self):
        ...

    @abstractmethod
    def _platform_cache(self):
        ...

    @abstractmethod
    def _follow_symlinks(self):
        ...

    @abstractmethod
    def _syscmd_ver(self):
        ...

    @abstractmethod
    def _norm_version(self):
        ...

    @abstractmethod
    def _uname_cache(self):
        ...

    @abstractmethod
    def _node(self):
        ...

    @abstractmethod
    def _syscmd_file(self):
        ...


class AbcPlatform(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mplatform"

    __doc__ = _interface.local_platform.__doc__

    os = mimportlib.module_property("os")
    sys = mimportlib.module_property("sys")

    __delegated__ = {_interface: _interface.local_platform}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_platform, self)

        mod._syscmd_ver = self._myrrh_syscmd_ver
        mod.java_ver = self._myrrh_java_ver
        mod._node = self._myrrh_node
        mod._syscmd_file = self._myrrh_syscmd_file
        mod._Processor.from_subprocess = self._myrrh_from_subprocess

        self.__delegate__(_interface, mod)

    def _myrrh_syscmd_ver(
        self,
        system="",
        release="",
        version="",
        supported_platforms=("win32", "win16", "dos"),
    ):
        """Tries to figure out the OS version used and returns
        a tuple (system, release, version).

        It uses the "ver" shell command for this which is known
        to exists on Windows, DOS. XXX Others too ?

        In case this fails, the given parameters are used as
        defaults.

        """
        sys = self.sys
        _ver_output = self._ver_output
        _norm_version = self._norm_version

        if sys.platform not in supported_platforms:
            return system, release, version

        # Try some common cmd strings
        for cmd in (b"ver", b"command /c ver", b'cmd /c "ver"'):
            # try:
            #    info = subprocess.check_output(cmd,
            #                                stdin=subprocess.DEVNULL,
            #                                stderr=subprocess.DEVNULL,
            #                                text=True,
            #                                encoding="locale",
            #                                shell=True)
            # except (OSError, subprocess.CalledProcessError) as why:
            #    #print('Command %s failed: %s' % (cmd, why))
            #    continue
            # else:
            #    break
            info, e, r = self.myrrh_os.cmd(cmd)
            if not r:
                break
        else:
            return system, release, version

        # Parse the output
        info = info.strip()
        m = _ver_output.match(info)
        if m is not None:
            system, release, version = m.groups()
            # Strip trailing dots from version and release
            if release[-1] == ".":
                release = release[:-1]
            if version[-1] == ".":
                version = version[:-1]
            # Normalize the version and build strings (eliminating additional
            # zeros)
            version = _norm_version(version)
        return system, release, version

    def _myrrh_java_ver(self, release="", vendor="", vminfo=("", "", ""), osinfo=("", "", "")):
        """Version interface for Jython.

        Returns a tuple (release, vendor, vminfo, osinfo) with vminfo being
        a tuple (vm_name, vm_release, vm_vendor) and osinfo being a
        tuple (os_name, os_version, os_arch).

        Values which cannot be determined are set to the defaults
        given as parameters (which all default to '').

        """
        import re

        _, e, _ = self.myrrh_os.cmd(b"%(java)s -XshowSettings:properties", java=b"java")

        props = re.findall(r"([.\w]+) = ([\S ]+)", e)
        if not props:
            return release, vendor, vminfo, osinfo

        props = dict(props)

        vendor = props.get("java.vendor", vendor)
        release = props.get("java.version", release)

        vm_name, vm_release, vm_vendor = vminfo
        vm_name = props.get("java.vm.name", vm_name)
        vm_vendor = props.get("java.vm.vendor", vm_vendor)
        vm_release = props.get("java.vm.version", vm_release)
        vminfo = vm_name, vm_release, vm_vendor

        os_name, os_version, os_arch = osinfo
        os_arch = props.get("java.os.arch", os_arch)
        os_name = props.get("java.os.name", os_name)
        os_version = props.get("java.os.version", os_version)
        osinfo = os_name, os_version, os_arch

        return release, vendor, vminfo, osinfo

    def _myrrh_node(self, default=""):
        o, _, r = self.myrrh_os.cmd(b"%(hostname)s", hostname=b"hostname")
        if r:
            return default
        return o.strip()

    def _myrrh_syscmd_file(self, target, default=""):
        """Interface to the system's file command.

        The function uses the -b option of the file command to have it
        omit the filename in its output. Follow the symlinks. It returns
        default in case the command should fail.

        """
        sys = self.sys
        _follow_symlinks = self._follow_symlinks
        os = self.os

        if sys.platform in ("dos", "win32", "win16"):
            # XXX Others too ?
            return default

        # try:
        #    subprocess = self.subprocess
        # except (ImportError):
        #    return default
        target = _follow_symlinks(target)
        # "file" output is locale dependent: force the usage of the C locale
        # to get deterministic behavior.
        env = dict(os.environ, LC_ALL="C")
        # try:
        #    # -b: do not prepend filenames to output lines (brief mode)
        #    output = subprocess.check_output(['file', '-b', target],
        #                                    stderr=subprocess.DEVNULL,
        #                                    env=env)
        # except (OSError, subprocess.CalledProcessError):
        #    return default
        output, _, rval = self.myrrh_os.cmd(
            b"%(file)s -b %(target)s",
            file=b"file",
            target=self.myrrh_os.fsencode(target),
            execute_env=env,
        )
        if not output or rval:
            return default
        # With the C locale, the output should be mostly ASCII-compatible.
        # Decode from Latin-1 to prevent Unicode decode error.
        return output.decode("latin-1")

    def _myrrh_from_subprocess(self):
        """
        Fall back to `uname -p`
        """
        o, _, r = self.myrrh_os.cmd(b"%(uname)s", uname=b"uname")
        if r:
            return None
        return o.strip()



import functools
import importlib.util
import warnings

from abc import ABC, abstractmethod
from myrrh.utils.mstring import typebytes, cast, typestr

from ...interfaces import (
    IMyrrhOs,
    ISystem,
    ICoreService,
    ICoreShellService,
    ICoreFileSystemService,
    ICoreStreamService,
    ABCDelegation,
)
from ....provider import Protocol

from ..objects import MyrrhEnviron
from ..managers import RuntimeCache, init_cache, runtime_cached_property, Acquiring
from ._syscall import RuntimeSyscall

__all__ = ("AbcMyrrhOs", "AbcRuntime", "AbcRuntimeDelegate")


def _validate_exe_args_values(command, working_dir, env):
    if env:
        for k, v in env.items():
            if b"\x00" in k or b"\x00" in v:
                raise ValueError("invalid environment value for %s: %s" % (k, v))
            if b"=" in k:
                raise ValueError("invalid environment key %s" % k)
    if not command:
        raise ValueError("command parameter must not be empty")

    if b"\x00" in command:
        raise ValueError("Invalid \\x00 not allowed in command")

    if working_dir and b"\x00" in working_dir:
        raise ValueError("Invalid \\x00 not allowed in working directory")


class AbcMyrrhOs(IMyrrhOs, ABCDelegation):
    @property
    @abstractmethod
    def _curdirb_(self):
        ...

    @property
    @abstractmethod
    def _pardirb_(self):
        ...

    @property
    @abstractmethod
    def _extsepb_(self):
        ...

    @property
    @abstractmethod
    def _sepb_(self):
        ...

    @property
    @abstractmethod
    def _pathsepb_(self):
        ...

    @property
    @abstractmethod
    def _altsepb_(self):
        ...

    @property
    @abstractmethod
    def _defpathb_(self):
        ...

    @property
    @abstractmethod
    def _devnullb_(self):
        ...

    @property
    @abstractmethod
    def _linesepb_(self):
        ...

    @abstractmethod
    def _getbinb_(self):
        ...

    @abstractmethod
    def _envb_(self):
        ...

    @abstractmethod
    def _getcwdb_(self):
        ...

    @abstractmethod
    def _gettmpdirb_(self):
        ...

    @abstractmethod
    def _localecode_(self):
        ...

    @abstractmethod
    def _fsencoding_(self):
        ...

    @abstractmethod
    def _fsencodeerrors_(self):
        ...

    @abstractmethod
    def _getdefaultencoding_(self):
        ...

    @abstractmethod
    def _getdefaultshellb_(self):
        ...

    @abstractmethod
    def _getdefaultshellargsb_(self):
        ...

    @abstractmethod
    def _getreadonlyenvb_(self):
        ...

    __delegated__ = (ISystem,)

    def __init__(self, system: ISystem):
        self.__delegate__(ISystem, system)

    @functools.cached_property
    def shell(self) -> ICoreShellService:
        return _RuntimeShell(self._delegate_.shell, self)  # type: ignore[attr-defined]

    @functools.cached_property
    def fs(self) -> ICoreFileSystemService:
        return _RuntimeFs(self._delegate_.fs, self)  # type: ignore[attr-defined]

    @functools.cached_property
    def stream(self) -> ICoreStreamService:
        return _RuntimeStream(self._delegate_.stream, self)  # type: ignore[attr-defined]

    def Stream(self, protocol: str | Protocol | None = None) -> ICoreStreamService:
        return _RuntimeStream(self._delegate_.Stream(protocol), self)

    def Fs(self, protocol: str | Protocol | None = None) -> ICoreFileSystemService:
        return _RuntimeFs(self._delegate_.Fs(protocol), self)

    def Shell(self, protocol: str | Protocol | None = None) -> ICoreShellService:
        return _RuntimeShell(self._delegate_.Shell(protocol), self)

    def syspathb(self, path):
        return self.normpath(self.joinpath(self.cwdb or b"", path))

    def getshellscriptb(self, script):
        return script

    curdirb = runtime_cached_property("curdirb")("_curdirb_")  # type: ignore[assignment]
    pardirb = runtime_cached_property("pardirb")("_pardirb_")  # type: ignore[assignment]
    extsepb = runtime_cached_property("extsepb")("_extsepb_")  # type: ignore[assignment]
    sepb = runtime_cached_property("sepb")("_sepb_")  # type: ignore[assignment]
    pathsepb = runtime_cached_property("pathsepb")("_pathsepb_")  # type: ignore[assignment]
    altsepb = runtime_cached_property("altsepb")("_altsepb_")  # type: ignore[assignment]
    linesepb = runtime_cached_property("linesepb")("_linesepb_")  # type: ignore[assignment]
    defpathb = runtime_cached_property("defpathb", init_cfg_path="system.defpathb")("_defpathb_")  # type: ignore[assignment]
    devnullb = runtime_cached_property("devnullb", init_cfg_path="system.devnullb")("_devnullb_")  # type: ignore[assignment]

    @runtime_cached_property("_shellb", init_cfg_path="system.shellb")
    def shellb(self):
        return self._getdefaultshellb_()

    @runtime_cached_property("_shellargsb", init_cfg_path="system.shellargsb")
    def shellargsb(self):
        return self._getdefaultshellargsb_()

    @runtime_cached_property("envb", init_value=dict(), init_cfg_path="system.envb")
    def envb(self):
        new_env = MyrrhEnviron({}, conv=self.fsencode, keyformat=self.environkeyformat)

        try:
            environ = self._envb_()
        except OSError:
            return new_env

        new_env.update(environ)

        return new_env

    @runtime_cached_property("cwdb", init_cfg_path="system.cwdb")
    def cwdb(self):
        return self._getcwdb_()

    @runtime_cached_property("tmpdirb", init_cfg_path="system.tmpdirb")
    def tmpdirb(self):
        return self._gettmpdirb_()

    @runtime_cached_property("rdenvb", init_cfg_path="system.rdenvb")
    def rdenvb(self):
        return self._getreadonlyenvb_()

    @runtime_cached_property("encoding", init_cfg_path="system.encoding")
    def defaultencoding(self):
        return self._getdefaultencoding_()

    @runtime_cached_property("fsencoding", init_cfg_path="system.fsencoding")
    def fsencoding(self):
        return self._fsencoding_()

    @runtime_cached_property("fsencodeerrors", init_cfg_path="system.fsencodeerrors")
    def fsencodeerrors(self):
        return self._fsencodeerrors_()

    @runtime_cached_property("localecode", init_cfg_path="system.localecode")
    def localcode(self):
        return self._localecode_()

    @runtime_cached_property("binb", init_cfg_path="system.binb")
    def getbinb(self):
        return self._getbinb_()

    @runtime_cached_property("modules", init_at_creation_time=True)
    def modules(self):
        return dict()

    @runtime_cached_property("impls", init_at_creation_time=True)
    def impls(self):
        return dict()

    def cmd(self, cmdline, **kwargs):
        out, err, rval = self.cmdb(cmdline, **kwargs)
        return self.shdecode(out), self.shdecode(err), rval

    def cmdb(self, cmdline, **kwargs):
        try:
            kwargs = {k.encode(): v for k, v in kwargs.items()}
            kwargs.update(self.getbinb)
            cmdline = cmdline % kwargs
        except KeyError as k:
            if not self.getbinb:
                raise OSError("Usable binaries list is empty, provider connection too long or failure?")

            raise OSError("%s not found on system" % k)

        working_dir = kwargs.get(b"execute_working_dir", None)
        env = kwargs.get(b"execute_env", None)
        out, err, rval = self.shell.execute(cmdline, working_dir=working_dir, env=env)

        return out.strip(), err.strip(), rval

    def shdecode(self, b, errors="surrogateescape") -> str:
        if not isinstance(b, bytes):
            return b
        return b.decode(encoding=self.defaultencoding, errors=errors)

    def shencode(self, s, errors="surrogateescape") -> bytes:
        if not isinstance(s, str):
            return s
        return s.encode(encoding=self.defaultencoding, errors=errors)

    def fsdecode(self, val) -> str:
        return typestr(encoding=self.fsencoding, errors=self.fsencodeerrors)(val)

    def fsencode(self, val) -> bytes:
        return typebytes(encoding=self.fsencoding, errors=self.fsencodeerrors)(val)

    def fdcast(self, val):
        try:
            if isinstance(val, int):
                val = str()
            return cast(val, encoding=self.fsencoding, errors=self.fsencodeerrors)

        except TypeError:
            return val.__class__

    def fscast(self, val):
        _type = self.fdcast(val)
        if issubclass(_type, self.PathLike):
            return cast(self._fspath_(val), encoding=self.fsencoding)
        return _type

    def shcast(self, val):
        return cast(val, encoding=self.defaultencoding)

    def getdefaultlocale(self):
        return self.localcode, self.defaultencoding

    def getdefaultshellb(self, args=()):
        return [self.shellb, *self.formatshellargs(args, defaultargs=self.shellargsb)] if self.shellb != b"" else args

    def getpathb(self, path=None):
        if path is None:
            return self.cwdb

        path = self.fsencode(path)

        if self.isabs(path):
            return self.normpath(path)

        return self.syspathb(path)

    def getenvb(self, env=None):
        if env is not None:
            _env = MyrrhEnviron({}, conv=self.fsencode)
            _env.update(env)
            return _env

        return self.envb

    def getpath(self, path=None):
        return self.fsdecode(self.getpathb(self.fsencode(path)))

    def getenv(self):
        return MyrrhEnviron(self.getenvb(), conv=self.fsdecode, keyformat=self.environkeyformat)

    def setenvb(self, env):
        conv = self.fsencode
        new_env = MyrrhEnviron({}, conv=conv, keyformat=self.environkeyformat)
        new_env.update(env)
        self.envb = new_env

    from os import PathLike, fspath  # type: ignore[misc]

    _fspath_ = staticmethod(fspath)

    def f(self, path, *, dir_fd=None):
        if isinstance(path, int):
            return self._get_runtime_object_(path).info

        return self.p(path, dir_fd=dir_fd)

    def p(self, path, *, dir_fd=None):
        try:
            path = self._fspath_(path)

            if isinstance(path, str):
                path = self.fsencode(path)

            if dir_fd is not None:
                path = self.join(self.f(dir_fd), path)

            return path

        except TypeError:
            if isinstance(path, (bytearray, memoryview)):
                warnings.warn(
                    "%s type deprecated for path parameter" % path.__class__.__name__,
                    DeprecationWarning,
                )
                return self.p(bytes(path), dir_fd=dir_fd)

        raise TypeError("path should be string, bytes or os.PathLike, not {}".format(path.__class__.__name__))


class _RuntimeShell(ICoreShellService, ICoreService, ABCDelegation):
    __delegated__ = (ICoreShellService, ICoreService)

    def __init__(self, shell, runtime: AbcMyrrhOs):
        self.__delegate__(ICoreShellService, shell)

        self._runtime = runtime

    def _getenv(self, env):
        env = self._runtime.getenvb(env)
        if env is None:
            return
        env = dict(env)
        for k in self._runtime.rdenvb:
            env.pop(k, None)
        return dict(env)  # need to be a dict

    def execute(self, command, working_dir=None, env=None, *, extras=None):
        _validate_exe_args_values(command, working_dir, env)

        try:
            working_dir = self._runtime.getpathb(working_dir)
        except Acquiring:
            working_dir = None
        try:
            env = self._getenv(env)
        except Acquiring:
            env = None
        try:
            command = self._runtime.getshellscriptb(command)
        except Acquiring:
            pass

        return self._delegate_.execute(command, working_dir, env, extras=extras)

    def spawn(self, command, working_dir=None, env=None, extras=None):
        _validate_exe_args_values(command, working_dir, env)

        working_dir = self._runtime.getpathb(working_dir)
        env = self._getenv(env)
        return self._delegate_.spawn(command, working_dir, env, extras=extras)


class _RuntimeFs(ICoreFileSystemService, ICoreService, ABCDelegation):
    __delegated__ = (ICoreFileSystemService, ICoreService)

    def __init__(self, fs: ICoreFileSystemService, runtime: AbcMyrrhOs):
        self.__delegate__(ICoreFileSystemService, fs)

        self._runtime: AbcMyrrhOs = runtime

    def rm(self, file_path, *, extras=None):
        return self._delegate_.rm(self._runtime.getpathb(file_path), extras=extras)

    def mkdir(self, path, *, extras=None):
        return self._delegate_.mkdir(self._runtime.getpathb(path), extras=extras)

    def rmdir(self, path, *, extras=None):
        return self._delegate_.rmdir(self._runtime.getpathb(path), extras=extras)

    def is_container(self, path, *, extras=None):
        return self._delegate_.is_container(self._runtime.getpathb(path), extras=extras)

    def exist(self, path, *, extras=None):
        return self._delegate_.exist(self._runtime.getpathb(path), extras=extras)

    def list(self, path, *, extras=None):
        return self._delegate_.list(self._runtime.getpathb(path), extras=extras)

    def stat(self, path, *, extras=None):
        return self._delegate_.stat(self._runtime.getpathb(path), extras=extras)


class _RuntimeStream(ICoreStreamService, ABCDelegation):
    __delegated__ = (ICoreStreamService, ICoreService)

    def __init__(self, stream: ICoreStreamService, runtime: IMyrrhOs):
        self.__delegate__(ICoreStreamService, stream)
        self._runtime = runtime

    def open_file(self, path: bytes, wiring: int, *, extras: dict | None = None) -> tuple[bytes, int]:
        return self._delegate_.open_file(self._runtime.getpathb(path), wiring=wiring, extras=extras)

    def open_process(
        self,
        path: bytes,
        wiring: int,
        args: list[bytes],
        working_dir: bytes | None = None,
        env: dict[bytes, bytes] | None = None,
        *,
        extras: dict | None = None,
    ) -> tuple[bytes, int, int, int, int]:
        return self._delegate_.open_process(
            self._runtime.getpathb(path),
            wiring=wiring,
            args=args,
            working_dir=working_dir,
            env=env,
            extras=extras,
        )


class AbcRuntime(ABC):
    __frameworkpath__: str
    __root_system__: ISystem

    def __new__(cls, system: ISystem):
        system = getattr(system, "__root_system__", system)

        with system.lock:
            __frameworkpath__ = getattr(cls, "__frameworkpath__", None)
            cache: RuntimeCache = getattr(system, "__m_runtime_cache__", None)  # type: ignore[assignment]

            if __frameworkpath__ and cache:
                impls = cache.get("impls") or tuple()

                if __frameworkpath__ in impls:
                    return impls[__frameworkpath__]

            clsname = cls.__name__

            myrrh_os = getattr(system, "__m_runtime_os__", None)
            myrrh_syscall = getattr(system, "__m_runtime_syscall__", None)

            arch = system.cfg.system.os

            if not arch:
                raise TypeError('no system type specified for %s, required "posix" or "nt"' % system.cfg.id)

            if not myrrh_os:
                modsys = importlib.import_module(f"myrrh.framework.arch.{arch}.myrrhos")
                myrrhos_cls = getattr(modsys, "MyrrhOs")
                myrrh_os = myrrhos_cls(system)

                myrrh_syscall = RuntimeSyscall(myrrh_os)
                myrrh_cache = RuntimeCache()

                init_cache(myrrh_cache, myrrh_os)
                init_cache(myrrh_cache, myrrh_syscall)

                setattr(system, "__m_runtime_os__", myrrh_os)
                setattr(system, "__m_runtime_syscall__", myrrh_syscall)
                setattr(system, "__m_runtime_cache__", myrrh_cache)

            if clsname.startswith("Abc"):
                try:
                    spec = importlib.util.find_spec(f"myrrh.framework.arch.{arch}.{cls.__frameworkpath__}")
                except ImportError:
                    pass
                else:
                    if spec:
                        mod = importlib.import_module(f"myrrh.framework.arch.{arch}.{cls.__frameworkpath__}")
                        clsname = getattr(mod, "__mlib__", None) or clsname
                        cls = getattr(mod, clsname)

            clsname = clsname[3:] if (clsname.startswith("Abc")) else clsname

            concrete_cls = type(
                clsname,
                (cls,),
                {
                    "__root_system__": system,
                    "myrrh_os": myrrh_os,
                    "myrrh_syscall": myrrh_syscall,
                    "__new__": ABC.__new__,
                },
            )

            self = concrete_cls()

            frameworkpath = getattr(self, "__frameworkpath__", None)
            if frameworkpath:
                try:
                    assert frameworkpath not in self.myrrh_os.impls
                except AssertionError:
                    pass
                self.myrrh_os.impls[frameworkpath] = self

            return self

    def __del__(self):
        cache = getattr(self.__root_system__, "__m_runtime_cache__", None)
        frameworkpath = getattr(self, "__frameworkpath__", None)

        if cache and frameworkpath:
            cache["impls"].pop(self.__frameworkpath__, None)

    @property
    @abstractmethod
    def myrrh_os(self) -> AbcMyrrhOs:
        ...

    @property
    @abstractmethod
    def myrrh_syscall(self) -> RuntimeSyscall:
        ...


class AbcRuntimeDelegate(AbcRuntime, ABCDelegation):
    ...



import functools
import typing

import mlib

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate, _mlib_

__mlib__ = "AbcImportLib"


def module_property(module, default=None):
    def get_str(self):
        importlib = AbcImportLib(self)
        try:
            return importlib.import_module(module)
        except ImportError:
            if default is not None:
                return default
            raise

    def get_mod(self):
        try:
            Abc = _mlib_(module)
            return Abc(self)
        except ImportError:
            if default is not None:
                return default
            raise

    if isinstance(module, str):
        return functools.cached_property(get_str)

    return functools.cached_property(get_mod)


class _interface(ABC):
    import importlib as local_importlib
    import importlib.util

    @abstractmethod
    def invalidate_caches(self) -> None:
        ...

    @abstractmethod
    def find_loader(name, path=None) -> typing.Any:
        ...

    @abstractmethod
    def import_module(name, package=None) -> typing.Any:
        ...

    @abstractmethod
    def reload(module) -> typing.Any:
        ...

    @abstractmethod
    def __import__(name, globals=None, locals=None, fromlist=(), level=0) -> typing.Any:
        ...


class AbcImportLib(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mimportlib"

    __all__ = ["invalidate_caches", "find_loader", "reload", "__import__"]

    __doc__ = _interface.local_importlib.__doc__

    __delegated__ = {_interface: _interface.local_importlib}
    __delegate_check_type__ = False

    sys = module_property("sys")

    def __init__(self, *a, **kwa):
        mod = _interface.importlib.util.module_from_spec(self.local_importlib.__spec__)
        assert mod is not self.local_importlib
        mod.__loader__.exec_module(mod)

        mod.import_module = self._myrrh_import_module
        mod.__import__ = self._myrrh__import__

        self.__delegate__(_interface, mod)

    def _myrrh_import_module(self, name, package=None):
        package = f"mlib.py.{package}" if package else "mlib.py"

        mod = self.local_importlib.import_module(f".{name}", package)

        try:
            return mod._mlib_cls(self)
        except (NotImplementedError, TypeError, AttributeError):
            raise ImportError

    def _myrrh__import__(self, name, globals=None, locals=None, fromlist=(), level=0):
        if level:
            globals_ = globals if globals is not None else {}
            package = self.local_importlib._bootstrap._calc___package__(globals_)
            name = self.local_importlib._bootstrap._resolve_name(name, package, level)

        if not fromlist:
            mod_name = name.partition(".")[0]
        else:
            mod_name = name

        if self.sys.modules.get(name) is None:
            mlib_name = f"mlib.py.{name}" if name else "mlib.py"

            with mlib.mlib_select(self):
                _mod = self.local_importlib.__import__(mlib_name, globals, locals, fromlist, 0)

                if not fromlist:
                    mod = getattr(_mod.py._mlib_cls(self), mod_name)
                else:
                    mod = _mod._mlib_cls(self)

                try:
                    new_mod = _mlib_(mod, lambda s: mod)(self)
                    if getattr(new_mod, "__name__", "").startswith("myrrh.framework.mpython"):
                        new_mod = mlib.__class__(
                            mod_name,
                            module=new_mod,
                            package=".".join((mlib.py.__package__, mod_name)),
                            parent=mlib.py,
                        )
                    self.sys.modules[mod_name] = new_mod
                except (NotImplementedError, TypeError, AttributeError) as e:
                    raise ImportError(e)

                mod = mlib.py
                fulln = ""
                for n in name.split("."):
                    fulln = fulln + n
                    mod = getattr(_mlib_(mod, mod), n)
                    if self.sys.modules.get(fulln) is None:
                        new_mod = _mlib_(mod, lambda s: mod)(self)
                        if getattr(new_mod, "__name__", "").startswith(mlib.py.__package__):
                            new_mod = mlib.__class__(
                                mod_name,
                                module=new_mod,
                                package=".".join((mlib.py.__package__, mod_name)),
                                parent=mlib.py,
                            )

                        self.sys.modules[fulln] = new_mod
                    fulln = fulln + "."

        return self.sys.modules[mod_name]



# -*- coding: utf-8 -*-

"""
**Filemode python module (from python3.x) **

.. note:: this module should not be used outside Myrrh library.

This module implements filemode as specified in python3.x

-----------

"""

import stat

_filemode_table = (
    (
        (stat.S_IFLNK, "l"),
        (stat.S_IFREG, "-"),
        (stat.S_IFBLK, "b"),
        (stat.S_IFDIR, "d"),
        (stat.S_IFCHR, "c"),
        (stat.S_IFIFO, "p"),
    ),
    ((stat.S_IRUSR, "r"),),
    ((stat.S_IWUSR, "w"),),
    ((stat.S_IXUSR | stat.S_ISUID, "s"), (stat.S_ISUID, "S"), (stat.S_IXUSR, "x")),
    ((stat.S_IRGRP, "r"),),
    ((stat.S_IWGRP, "w"),),
    ((stat.S_IXGRP | stat.S_ISGID, "s"), (stat.S_ISGID, "S"), (stat.S_IXGRP, "x")),
    ((stat.S_IROTH, "r"),),
    ((stat.S_IWOTH, "w"),),
    ((stat.S_IXOTH | stat.S_ISVTX, "t"), (stat.S_ISVTX, "S"), (stat.S_IXOTH, "x")),
)


def filemode(mode):
    """
    Convert a file's mode to a string of the form '-rwxrwxrwx'.
    :param mode: stat mode as returned by stat

    :return: given mode formyrrhd string
    :rtype: str
    """
    perm = []
    for table in _filemode_table:
        for bit, char in table:
            if mode & bit == bit:
                perm.append(char)
                break
        else:
            perm.append("-")

    return "".join(perm)



import importlib
import os
import sys
import click
from click.decorators import pass_context
import pdoc
import pathlib
import webbrowser


from myrrh.core.services import cfg_get

from myrrh.tools.myrrhc_ext import myrrhc_cmds, cmd

from myrrh.tools.myrrhc.__doc__ import alias

pdoc.render.configure(docformat="restructuredtext")


def man_complete(_ctx, _cli, _incomplet):
    return list(alias())


@myrrhc_cmds.command()
@cmd.argument("alias_or_module_name", required=False, default="", shell_complete=man_complete)
@cmd.argument("sub_commands", required=False, nargs=-1)
@cmd.option(
    "-f",
    "--force-build",
    is_flag=True,
    required=False,
    default=cmd.option_cfg_default("man_always_rebuild", False),
)
@cmd.option(
    "-d",
    "--directory",
    required=False,
    default=cmd.option_cfg_default("man_dir", os.path.join(cfg_get(key="@etc@")[0], "man")),
)
@pass_context
def man(ctx: click.Context, alias_or_module_name, force_build, sub_commands, directory):
    """
    Display help information about a command or a python module.
    """
    _tmpdir = pathlib.Path(directory)

    if not alias_or_module_name:
        alias_or_module_name = "myrrh"

    module_name = alias().get(alias_or_module_name, alias_or_module_name)
    module_name, _, qualname = alias().get(alias_or_module_name, alias_or_module_name).partition("#")

    if qualname:
        alias_or_module_name = module_name
        module_name = alias().get(alias_or_module_name, alias_or_module_name)

    tmpdir = _tmpdir.joinpath(module_name)
    os.makedirs(tmpdir, exist_ok=True)
    index = tmpdir.joinpath("index.html")

    if not os.path.exists(index) or force_build:
        docs = {}
        try:
            mod = importlib.import_module(f"{module_name}.__doc__")

            if hasattr(mod, "__doc__"):
                o = type(sys)(alias_or_module_name)
                o.__doc__ = mod.__doc__() if callable(mod.__doc__) else mod.__doc__
                o.__doc__ = o.__doc__.replace("_main_", alias_or_module_name)
                docs[alias_or_module_name] = pdoc.doc.Module(o)
                docs[alias_or_module_name].is_package = True

            if hasattr(mod, "__docgen__"):
                for name, txt in mod.__docgen__().items():
                    o = type(sys)(name)
                    o.__doc__ = txt.replace("_main_", name.split(".")[-1])
                    docs[f"{alias_or_module_name}.{name}"] = pdoc.doc.Module(o)

        except (ModuleNotFoundError, RuntimeError):
            docs = {module_name: pdoc.doc.Module.from_name(module_name)}

        for name, doc in docs.items():
            out = pdoc.render.html_module(module=doc, all_modules=docs)
            outfile = tmpdir / f"{name.replace('.', '/')}.html"
            outfile.parent.mkdir(parents=True, exist_ok=True)
            outfile.write_bytes(out.encode())

        index = pdoc.render.html_index(docs)
        if index:
            (tmpdir / "index.html").write_bytes(index.encode())

        search = pdoc.render.search_index(docs)
        if search:
            (tmpdir / "search.js").write_bytes(search.encode())

    if qualname:
        webbrowser.open(str(tmpdir / f"{alias_or_module_name.replace('.', '/')}" / f"{'/'.join((qualname.replace('.', '/'), *sub_commands))}.html"))
        return

    webbrowser.open(str(tmpdir / "index.html"))



__all__ = ("MyrrhEnviron",)


class MyrrhEnviron:
    class _itemsview:
        def __init__(self, myrrh_environ):
            self.__myrrh_environ = myrrh_environ
            self.__it = iter(myrrh_environ)

        def __iter__(self):
            return self

        def __next__(self):
            k = next(self.__it)
            return k, self.__myrrh_environ[k]

    class _valuesview:
        def __init__(self, myrrh_environ):
            self.__myrrh_environ = myrrh_environ
            self.__it = iter(myrrh_environ)

        def __iter__(self):
            return self

        def __next__(self):
            return self.__myrrh_environ[next(self.__it)]

    class _keysview:
        def __init__(self, myrrh_environ):
            self.__myrrh_environ = myrrh_environ
            self.__it = iter(myrrh_environ._env)

        def __iter__(self):
            return self

        def __next__(self):
            result = next(self.__it)
            while result in self.__myrrh_environ._filter:
                result = next(self.__it)
            return self.__myrrh_environ._conv(result)

    def __init__(self, env, *, conv=None, keyformat=None, filter=[]):
        self.__env = env

        self._conv = conv or (lambda k: k)
        self._kfmt = (lambda k: self._conv(keyformat(k))) if keyformat else conv

        self.__filter = [self._kfmt(k) for k in filter]

    def __repr__(self):
        return "environ(%s)" % repr({k: v for k, v in self.items()})

    @property
    def _env(self):
        return self.__env

    @property
    def _filter(self):
        return self.__filter

    def __eq__(self, environ):
        return {k: v for k, v in self.items()} == {self._kfmt(k): self._conv(v) for k, v in environ.items()}

    def __setitem__(self, key, value):
        self._env[self._kfmt(key)] = value

    def __getitem__(self, key):
        try:
            return self._conv(self._env[self._kfmt(key)])
        except KeyError:
            raise KeyError(key) from None

    def __delitem__(self, key):
        try:
            del self._env[self._kfmt(key)]
        except KeyError:
            raise KeyError(key) from None

    def __iter__(self):
        return iter(self._keysview(self))

    def __contains__(self, key):
        return self._env.__contains__(self._kfmt(key))

    def __len__(self):
        return self._env.__len__()

    def __str__(self):
        return self._env.__str__()

    def clear(self):
        self._env.clear()

    def copy(self):
        return MyrrhEnviron(self._env.copy(), conv=self._conv, keyformat=self._kfmt, filter=self._filter)

    def get(self, key, default=None):
        return self._conv(self._env.get(self._kfmt(key), default))

    def items(self):
        return self._itemsview(self)

    def keys(self):
        return self._keysview(self)

    def values(self):
        return self._valuesview(self)

    def pop(self, key, *args, **kwargs):
        if len(kwargs) > 1 or len(args) > 1:
            raise ValueError("pop is called with too many arguments")

        try:
            default = args[0] if len(args) > 0 else kwargs.pop("default")
            return self._conv(self._env.pop(self._kfmt(key), self._conv(default)))
        except KeyError:
            return self._conv(self._env.pop(self._kfmt(key)))

    def popitem(self):
        k, v = self._env.popitem()
        return self._conv(k), self._conv(v)

    def setdefault(self, key, default=None):
        return self._conv(self._env.setdefault(self._kfmt(key), self._conv(default)))

    def update(self, other={}, **kwargs):
        kwargs.update(other)
        for k, v in kwargs.items():
            self._env.update({self._kfmt(k): self._conv(v)})



from myrrh.core.services.system import AbcRuntime

__mlib__ = "AbcIo"


class AbcIo(AbcRuntime):
    __frameworkpath__ = "mpython._mio"

    from _io import FileIO  # type: ignore

    class FileIO(FileIO):  # type: ignore[no-redef]
        def open(*a, **kwa):
            ...



# -*- coding: utf-8 -*-

import xmlrpc.client

import pickle
import traceback

import myrrh

__all__ = ["encode_arg", "encode_arg_self", "decode_arg", "decode_arg_self"]


def _except_encode(exc, trcbck):
    raise xmlrpc.client.Fault(0, pickle.dumps((exc, trcbck)))


def _except_decode(fault):
    try:
        exc, trcbck = pickle.loads(fault.faultString.data)
    except Exception as e:
        exc = fault
        trcbck = traceback.format_tb(e.__traceback__)

    exc.straceback = trcbck

    myrrh.log.debug(f'{"".join(trcbck)}{str(exc)}\n')

    raise exc


def _encode(val):
    return pickle.dumps(val)


def _decode(val):
    if hasattr(val, "data"):
        return pickle.loads(val.data)
    return val


def encode_arg(proc):
    def wrapper(self, *args, **kwargs):
        try:
            return _decode(proc(self, _encode(args), _encode(kwargs)))
        except xmlrpc.client.Fault as fault:
            _except_decode(fault)

    return wrapper


def encode_arg_self(proc):
    def wrapper(*args, **kwargs):
        try:
            return _decode(proc(_encode(args), _encode(kwargs)))
        except xmlrpc.client.Fault as fault:
            _except_decode(fault)

    return wrapper


def decode_arg(proc):
    def wrapper(self, args, kwargs):
        try:
            return _encode(proc(self, *_decode(args), **_decode(kwargs)))
        except Exception as e:
            _except_encode(e, trcbck=traceback.format_tb(e.__traceback__))

    return wrapper


def decode_arg_self(proc):
    def wrapper(args, kwargs):
        try:
            return _encode(proc(*_decode(args), **_decode(kwargs)))
        except Exception as e:
            _except_encode(e, trcbck=traceback.format_tb(e.__traceback__))

    return wrapper
